#include "component_text.h"

void _component_text_glyph_size_get(ComponentText* self, Texture* texture, vec2 size);
void _component_text_glyph_position_get(ComponentText* self, Texture* texture, vec3 position);
void _component_text_glyph_entities_init(ComponentText* self, ECS* ecs);
void _component_text_glyph_entities_free(ComponentText* self, ECS* ecs);
void _component_text_glyphs_position_set(ComponentText* self, ECS* ecs);
void _component_text_glyphs_color_set(ComponentText* self, ECS* ecs);

void
_component_text_glyph_size_get(ComponentText* self, Texture* texture, vec2 size)
{
	size[0] = texture->size[0] * self->scale[0];
	size[1] = texture->size[1] * self->scale[1];
}

void
_component_text_glyph_entities_init(ComponentText* self, ECS* ecs)
{
	vec2 dropShadowOffset;

	glm_vec2_zero(dropShadowOffset);

	dropShadowOffset[1] = (s32)(self->font->size * COMPONENT_TEXT_DROP_SHADOW_FONT_SIZE_MULTIPLIER);

	for (s32 i = 0; i < (s32)self->length; i++)
	{
		EntityID glyphID;
		vec2 glyphSize;
		vec3 glyphPosition;
		char glyph;

		glyph = self->string[i];

		if (glyph == '\n')
			glyph = ' ';

		_component_text_glyph_size_get(self, &self->font->glyphTextures[(s32)glyph], glyphSize);

		glyphID = entity_glyph_add
		(
			ecs,
			&self->font->glyphTextures[(s32)glyph],
			self->buffer,
			glyphSize,
			self->position,
			self->color,
			self->dropShadowColor,
			dropShadowOffset
		);

		vector_push(&self->glyphEntities, &glyphID);
	}

	_component_text_glyphs_position_set(self, ecs);

	for (s32 i = 0; i < (s32)self->length; i++)
	{
		EntityID* glyphEntityID;
		ComponentDropShadow* glyphEntityDropShadow;
		
		glyphEntityID = (EntityID*)vector_get(&self->glyphEntities, i);

		glyphEntityDropShadow = ecs_component_get(ecs, COMPONENT_DROP_SHADOW, *glyphEntityID);
		
		component_drop_shadow_tick(glyphEntityDropShadow, ecs);
	}
}

void
_component_text_glyph_entities_free(ComponentText* self, ECS* ecs)
{
	for (s32 i = 0; i < (s32)self->length; i++)
	{
		EntityID* glyphEntity;

		glyphEntity = (EntityID*)vector_get(&self->glyphEntities, i);
		
		ecs_entity_delete(ecs, *glyphEntity);
	}

	vector_clear(&self->glyphEntities);
}

void
_component_text_glyphs_position_set(ComponentText* self, ECS* ecs)
{
	GlyphMetrics glyphMetrics;
	f32 currentLineWidth;
	f32 increment;
	f32 kerning;
	f32 lineBegin;
	f32 lineSkip;
	f32 advance;
	vec3 position;

	lineSkip = font_line_skip_get(self->font);

	glm_vec3_copy(self->position, position);

	currentLineWidth = 0.0f;
	lineBegin = position[0];

	for (s32 i = 0; i < (s32)self->length; i++)
	{
		EntityID* glyphEntityID;
		ComponentTextureQuad* glyphEntityTextureQuad;
		char glyph;

		glyph = self->string[i];
		glyphEntityID = (EntityID*)vector_get(&self->glyphEntities, i);
		glyphEntityTextureQuad = ecs_component_get(ecs, COMPONENT_TEXTURE_QUAD, *glyphEntityID);

		font_glyph_metrics_init(self->font, &glyphMetrics, glyph);

		advance = glyphMetrics.advance;

		if 
		(
			glyph == '\n' ||
			(glyph == ' ' && (currentLineWidth > self->wrap) && self->wrap != -1)
		)
		{
			position[0] = lineBegin;
			position[1] += lineSkip * self->scale[1];

			advance = 0.0f;

			currentLineWidth = 0.0f;
		}

		if (i < (s32)self->length)
			kerning = (f32)font_glyph_kerning_get(self->font, glyph, self->string[i + 1]);
		else
			kerning = 0.0f;

		glm_vec3_copy(position, glyphEntityTextureQuad->position);

		increment = advance + kerning;

		position[0] += increment;
		position[2] -= COMPONENT_TEXT_Z_OFFSET_INCREMENT;

		currentLineWidth += increment;
	}
}

void
_component_text_glyphs_color_set(ComponentText* self, ECS* ecs)
{
	for (s32 i = 0; i < (s32)self->length; i++)
	{
		EntityID* glyphEntityID;
		ComponentTextureQuad* glyphEntityTextureQuad;
		ComponentDropShadow* glyphEntityDropShadow;
		ComponentTextureQuad* glyphEntityDropShadowTextureQuad;

		glyphEntityID = (EntityID*)vector_get(&self->glyphEntities, i);

		glyphEntityTextureQuad = ecs_component_get(ecs, COMPONENT_TEXTURE_QUAD, *glyphEntityID);
		glyphEntityDropShadow = ecs_component_get(ecs, COMPONENT_DROP_SHADOW, *glyphEntityID);
		glyphEntityDropShadowTextureQuad = ecs_component_get(ecs, COMPONENT_TEXTURE_QUAD, glyphEntityDropShadow->spriteID);

		glm_vec4_copy(self->color, glyphEntityTextureQuad->color);
		glm_vec4_copy(self->colorCover, glyphEntityTextureQuad->colorCover);

		glyphEntityDropShadowTextureQuad->color[3] = self->color[3];
		glm_vec4_copy(self->colorCover, glyphEntityDropShadowTextureQuad->colorCover);
	}
}

void
component_text_string_set(ComponentText* self, ECS* ecs, char* string, u32 length)
{
	_component_text_glyph_entities_free(self, ecs);

	if (length > COMPONENT_TEXT_STRING_MAX)
		return;

	self->length = length;

	memset(self->string, '\0', sizeof(char) * COMPONENT_TEXT_STRING_MAX);
	memcpy(self->string, string, sizeof(char) * self->length);

	_component_text_glyph_entities_init(self, ecs);
}

void
component_text_add(ComponentText* self, ECS* ecs)
{
	glm_vec4_copy(COMPONENT_TEXT_COLOR_DEFAULT, self->color);
	glm_vec4_copy(COMPONENT_TEXT_COLOR_COVER_DEFAULT, self->color);
	glm_vec4_copy(COMPONENT_TEXT_DROP_SHADOW_COLOR_DEFAULT, self->dropShadowColor);
	glm_vec2_copy(COMPONENT_TEXT_SCALE_DEFAULT, self->scale);

	self->wrap = COMPONENT_TEXT_WRAP_DEFAULT;

	vector_init(&self->glyphEntities, sizeof(EntityID));
}

void
component_text_delete(ComponentText* self, ECS* ecs)
{
	_component_text_glyph_entities_free(self, ecs);

	vector_free(&self->glyphEntities);
}

void
component_text_tick(ComponentText* self, ECS* ecs)
{
	_component_text_glyphs_position_set(self, ecs);
	_component_text_glyphs_color_set(self, ecs);
}

void
component_text_init
(
	ComponentText* self, 
	ECS* ecs,
	Font* font,
	RendererBuffer buffer,
	char* string,
	vec3 position,
	u32 length,
	s32 wrap
)
{
	self->font = font;
	self->wrap = wrap;
	self->buffer = buffer;

	glm_vec3_copy(position, self->position);

	component_text_string_set(self, ecs, string, length);
}
